use crate::func::*;

struct LazyClass;

impl WeakFunctor for LazyClass {
    type F<'a, A: 'a> = Box<dyn 'a + FnOnce() -> A>;
}

impl Functor for LazyClass {
    fn fmap<'a, A: 'a, B: 'a>(f: impl 'a + FnOnce(A) -> B, fa: Self::F<'a, A>) -> Self::F<'a, B> {
        Box::new(|| f(fa()))
    }

    fn replace<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, b: B) -> Self::F<'a, B> {
        drop(fa);
        Box::new(|| b)
    }

    fn void<'a, A: 'a>(fa: Self::F<'a, A>) -> Self::F<'a, ()> {
        drop(fa);
        Box::new(|| ())
    }
}

impl ApplicativeSeq for LazyClass {
    fn seq<'a, A: 'a, B: 'a>(
        ff: Self::F<'a, impl 'a + FnOnce(A) -> B>,
        fa: Self::F<'a, A>,
    ) -> Self::F<'a, B> {
        Box::new(|| ff()(fa()))
    }
}

impl ApplicativeLA2 for LazyClass {
    fn la2<'a, A: 'a, B: 'a, C: 'a>(
        f: impl 'a + FnOnce(A, B) -> C,
        fa: Self::F<'a, A>,
        fb: Self::F<'a, B>,
    ) -> Self::F<'a, C> {
        Box::new(|| f(fa(), fb()))
    }
}

impl ApplicativeTuple for LazyClass {
    fn tuple<'a, A: 'a, B: 'a>((fa, fb): (Self::F<'a, A>, Self::F<'a, B>)) -> Self::F<'a, (A, B)>
    where
        Self: 'a,
    {
        Box::new(|| (fa(), fb()))
    }
}

impl Applicative for LazyClass {
    fn pure<'a, A: 'a>(a: A) -> Self::F<'a, A> {
        Box::new(|| a)
    }

    fn discard_first<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, B> {
        drop(fa);
        fb
    }

    fn discard_second<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, A> {
        drop(fb);
        fa
    }
}

impl Monad for LazyClass {
    fn bind<'a, A: 'a, B: 'a>(
        fa: Self::F<'a, A>,
        f: impl 'a + FnOnce(A) -> Self::F<'a, B>,
    ) -> Self::F<'a, B> {
        Box::new(|| f(fa())())
    }

    fn join<'a, A: 'a>(ffa: Self::F<'a, Self::F<'a, A>>) -> Self::F<'a, A> {
        Box::new(|| ffa()())
    }
}
